@@ -6,6 +6,8 @@ from calendar import calendar
 from django.db import models
 from django.db.models.query_utils import Q
 
+from django.db.models.signals import pre_delete, post_delete
+
 PERIOD_INFTY = 0 # в регистре нет периодичности и нет динамически хранящихся
 PERIOD_SECOND = 1 # до секунд
 PERIOD_MINUTE = 2
@@ -15,13 +17,14 @@ PERIOD_MONTH = 5
 PERIOD_QUARTER = 6
 PERIOD_YEAR = 7
 
-def normdate(period, date, begin = True):
-    '''
-    Метод нормализует дату в зависимости от периода
-    @ begin = True - даты выравниваются на начало, иначе на конец
-    '''
+def normdate(period, date, begin=True):
+    u""" Метод нормализует дату в зависимости от периода.
+
+    :param begin: True - даты выравниваются на начало, иначе на конец
+
+    """
     if not date:
-        return None        
+        return None
     if period == PERIOD_SECOND:
         return datetime(date.year, date.month, date.day, date.hour, date.minute, date.second)
     if period == PERIOD_MINUTE:
@@ -44,21 +47,23 @@ def normdate(period, date, begin = True):
         return datetime(date.year, 1 if begin else 12, 1 if begin else calendar.monthrange(date.year, 1 if begin else 12)[1], 0 if begin else 23, 0 if begin else 59, 0 if begin else 59)
     return date
 
-def shift_date(period, date, step = 1):
-    '''
-    Метод смещает дату на один пункт периода
-    @ step = смещение
-    '''
+
+def shift_date(period, date, step=1):
+    u""" Метод смещает дату на один пункт периода.
+
+    :param step: смещение
+
+    """
     if not date:
-        return None        
+        return None
     elif period == PERIOD_SECOND:
-        return date + timedelta(seconds = step)
+        return date + timedelta(seconds=step)
     elif period == PERIOD_MINUTE:
-        return date + timedelta(minutes = step)
+        return date + timedelta(minutes=step)
     elif period == PERIOD_HOUR:
-        return date + timedelta(hours = step)
+        return date + timedelta(hours=step)
     elif period == PERIOD_DAY:
-        return date + timedelta(days = step)
+        return date + timedelta(days=step)
     elif period == PERIOD_MONTH:
         #TODO: не знаю пока как быстро изменить дату на месяц, т.к. день может быть 31, а в пред. месяце его нет
         return date
@@ -70,20 +75,21 @@ def shift_date(period, date, step = 1):
     else:
         return date
 
+
+
 class OverlapError(Exception):
-    """
+    u"""
     Исключение возникающее при наложении интервалов друг на друга.
     """
     def __init__(self, cls=None):
         self.model_cls = cls
-#    def __str__(self):
-#        return self.exception_message
+
 
 class InvalidIntervalError(Exception):
-    '''
+    u"""
     Исключение возникающее при неверно заданом интервале,
     например если конечная дата меньше начальной.
-    '''
+    """
     pass
 
 class InfoModelMeta(models.base.ModelBase):
@@ -116,7 +122,7 @@ class InfoModelMeta(models.base.ModelBase):
                         new_field = models.DateField(db_index = field.db_index, default = default)
                         new_class.replace_field(field, new_field)
         # Если в модели определены специальные источники для полей, содержащих дату:
-        if hasattr(new_class, 'begin_src') and hasattr(new_class, 'end_src'):
+        if hasattr(new_class, 'begin_src') and hasattr(new_class, 'end_src') and new_class._meta.local_fields:
 
             def _get_fld(name):
                 # Возвращает поле из _meta по его name
@@ -128,6 +134,10 @@ class InfoModelMeta(models.base.ModelBase):
             if new_class.end_src != new_class.DEFAULT_END:
                 # Если в модели переопеределно поле даты конца, info_date_end нам не нужен. убираем его:
                 new_class._meta.local_fields.remove(_get_fld(new_class.DEFAULT_END))
+
+        # подключим сигналы к классу
+        new_class.connect_to_signals()
+
         return new_class
 
     def replace_field(new_class, old_field, new_field):
@@ -140,6 +150,62 @@ class InfoModelMeta(models.base.ModelBase):
         new_index = new_class._meta.local_fields.index(new_field)
         new_class._meta.local_fields.insert(index, new_class._meta.local_fields.pop(new_index))
 
+    def connect_to_signals(cls):
+        # подключимся к сигналам удаления моделей
+        pre_delete.connect(cls.infomodel_pre_delete, cls, dispatch_uid='base_info_model_pre_delete')
+        post_delete.connect(cls.infomodel_post_delete, cls, dispatch_uid='base_info_model_post_delete')
+
+    def infomodel_pre_delete(cls, instance, **kwargs):
+        # если в параметрах есть признак, то пропустим обработку сигналов
+        if instance._signal_params.get('skip_signals'):
+            return
+        # если объект уже хранится, то найдем записи для изменения
+        if instance.pk:
+            def get_rec(date, lookup):
+                # в зависимости от базовой модели разные фильтры
+                if isinstance(instance, BaseInfoModel):
+                    q = Q(info_date=date)
+                else:
+                    q = instance._gen_Q(lookup, '', date)
+                try:
+                    return instance.query_dimentions(instance).get(q)
+                except (
+                    instance.DoesNotExist,
+                    instance.MultipleObjectsReturned
+                ):
+                    return None
+            old_prev_rec = get_rec(instance.info_date_prev, 'end')
+            old_next_rec = get_rec(instance.info_date_next, 'end')
+        else:
+            old_prev_rec = None
+            old_next_rec = None
+        instance._signal_params['old_prev_rec'] = old_prev_rec
+        instance._signal_params['old_next_rec'] = old_next_rec
+
+    def infomodel_post_delete(cls, instance, **kwargs):
+        # если в параметрах есть признак, то пропустим обработку сигналов
+        if not instance._signal_params.get('skip_signals'):
+            # изменим найденные ранее записи
+            # TODO: тут можно оптимизировать сохранение,
+            # чтобы не было каскада сохранений
+            old_prev_rec = instance._signal_params.get('old_prev_rec')
+            old_next_rec = instance._signal_params.get('old_next_rec')
+            # нужно проверить что объект есть, т.к. он может быть удален
+            if old_prev_rec and cls.objects.filter(pk=old_prev_rec.pk).exists():
+                old_prev_rec.save()
+            if old_next_rec and cls.objects.filter(pk=old_next_rec.pk).exists():
+                old_next_rec.save()
+
+    @staticmethod
+    def suppress_signals(meth):
+        def inner(inst, *args, **kwargs):
+            inst._signal_params['skip_signals'] = True
+            res = meth(inst, *args, **kwargs)
+            inst._signal_params['skip_signals'] = False
+            return res
+        return inner
+
+
 class BaseInfoModel(models.Model):
     __metaclass__ = InfoModelMeta
 
@@ -148,7 +214,7 @@ class BaseInfoModel(models.Model):
     info_date_prev  = models.DateTimeField(db_index = True, default = datetime.min)
     info_date_next = models.DateTimeField(db_index = True, default = datetime.max)
     info_date = models.DateTimeField(db_index = True, blank = True)
-    
+
     dimentions = [] # перечень реквизитов-измерений
     period = PERIOD_DAY # тип периодичности
 
@@ -156,34 +222,53 @@ class BaseInfoModel(models.Model):
     MAX_DATE = datetime.max
     MIN_DATE = datetime.min
 
+    def __init__(self, *args, **kwargs):
+        super(BaseInfoModel, self).__init__(*args, **kwargs)
+        # словарь параметров для передачи через сигналы
+        self._signal_params = {}
+
     # сформируем запрос по ключу
     @classmethod
     def query_dimentions(cls, data):
-        '''
-        запрос по ключевым полям
-        @data - объект, который содержит данные для сохранения в регистр.
-                Это может быть либо 1) объект модели хранения данных регистра,
-                либо 2) словарь, ключами которого являются имена полей из модели хранения,
-                либо 3) объект любого класса, у которого есть атрибуты с именами
-        '''
-        # добавим в запрос фильтр по ключевым полям, со значениями = текущему объекту, переданным данным
+        """ Запрос по ключевым полям.
+
+        :param data: объект, который содержит данные для сохранения в
+            регистр. Это может быть:
+            - Либо объект модели хранения данных регистра,
+            - Либо словарь, ключами которого являются имена полей из
+              модели хранения,
+            - Либо объект любого класса, у которого есть атрибуты с
+              именами
+        """
+        # Добавим в запрос фильтр по ключевым полям, со значениями =
+        # текущему объекту, переданным данным
         query = cls.objects
         for dim_field in cls.dimentions:
             dim_val = None
             dim_attr = dim_field
+
             if isinstance(dim_field, models.Field):
                 dim_attr = dim_field.name
+
             if isinstance(data, BaseInfoModel):
+                if not hasattr(data, dim_attr):
+                    continue
                 dim_val = getattr(data, dim_attr, None)
+
             elif isinstance(data, dict):
-                if data.has_key(dim_attr):
-                    dim_val = data[dim_attr]
+                if dim_attr not in data:
+                    continue
+                dim_val = data[dim_attr]
+
             elif hasattr(data, dim_attr):
+                if not hasattr(data, dim_attr):
+                    continue
                 dim_val = getattr(data, dim_attr, None)
-            if dim_val:
-                query = query.filter(**{dim_attr: dim_val})
+
+            query = query.filter(**{dim_attr: dim_val})
+
         return query
-    
+
     # сформируем запрос данных на дату
     @classmethod
     def query_on_date(cls, data, date=None, next = False):
@@ -207,7 +292,7 @@ class BaseInfoModel(models.Model):
                 else:
                     query = query.filter(info_date__lte = q_date, info_date_next__gt = q_date)
         return query
-    
+
     # прямая запись объекта, чтобы можно было записывать без доп. обработки
     def _save(self, *args, **kwargs):
         super(BaseInfoModel, self).save(*args, **kwargs)
@@ -217,9 +302,9 @@ class BaseInfoModel(models.Model):
         # если не указана дата записи, то
         if self.period != PERIOD_INFTY and not self.info_date:
             raise Exception('Attribute info_date is not set!')
-        # нормализуем переданные дату и время 
+        # нормализуем переданные дату и время
         self.info_date = normdate(self.period,self.info_date)
-        
+
         # проверим на уникальность записи
         if self.period != PERIOD_INFTY:
             q = self.__class__.query_dimentions(self).filter(info_date = self.info_date)
@@ -254,7 +339,8 @@ class BaseInfoModel(models.Model):
         # сохраним запись
         self._save(*args, **kwargs)
         # изменим найденные ранее записи
-        #TODO: тут можно оптимизировать сохранение, чтобы небыло каскада сохранений
+        # TODO: тут можно оптимизировать сохранение,
+        # чтобы не было каскада сохранений
         if old_prev_rec:
             if old_prev_rec != new_prev_rec:
                 old_prev_rec.save()
@@ -271,44 +357,25 @@ class BaseInfoModel(models.Model):
                 new_next_rec._save()
 
     # прямое удаление, без обработки
+    @InfoModelMeta.suppress_signals
     def _delete(self, *args, **kwargs):
         super(BaseInfoModel, self).delete(*args, **kwargs)
 
-    # переопределим удаление объекта, чтобы изменять соседние записи
-    def delete(self, *args, **kwargs):
-        # если объект уже хранится, то найдем записи для изменения
-        if self.pk:
-            q = self.__class__.query_dimentions(self).filter(info_date = self.info_date_prev).exclude(pk = self.pk)
-            old_prev_rec = q.get() if len(q) == 1 else None
-            q = self.__class__.query_dimentions(self).filter(info_date = self.info_date_next).exclude(pk = self.pk)
-            old_next_rec = q.get() if len(q) == 1 else None
-        else:
-            old_prev_rec = None
-            old_next_rec = None
-        # удалим запись
-        self._delete(*args, **kwargs)
-        # изменим найденные ранее записи
-        #TODO: тут можно оптимизировать сохранение, чтобы небыло каскада сохранений
-        if old_prev_rec:
-            old_prev_rec.save()
-        if old_next_rec:
-            old_next_rec.save()
-    
     class Meta:
         abstract = True
 
 
 def RebuildInfoModel(cls):
-    '''
+    u"""
     Перестраивает связи между записями в регистре
-    ''' 
-    
+    """
+
     def eq_keys(dims, key1, key2):
         for key_attr in dims:
             if key1[key_attr] != key2[key_attr]:
                 return False
         return True
-    
+
     def get_key(dims, data):
         key = {}
         for key_attr in dims:
@@ -316,14 +383,15 @@ def RebuildInfoModel(cls):
             if isinstance(data, BaseInfoModel):
                 dim_val = getattr(data, key_attr, None)
             elif isinstance(data, dict):
-                if data.has_key(key_attr):
+                if key_attr in data:
                     dim_val = data[key_attr]
             elif hasattr(data, key_attr):
                 dim_val = getattr(data, key_attr, None)
             key[key_attr] = dim_val
         return key
-    
-    # вытащить записи регистра сгруппированные по ключевым полям и отсортированные по дате
+
+    # Вытащить записи регистра сгруппированные по ключевым полям и
+    # отсортированные по дате
     query = cls.objects
     order = []
     dims = []
@@ -363,6 +431,8 @@ def RebuildInfoModel(cls):
         last_rec.info_date_next = normdate(period, datetime.max)
         last_rec._save()
 
+
+
 class BaseIntervalInfoModel(models.Model):
     __metaclass__ = InfoModelMeta
     # Константы по умолчанию для источника даты начала и конца:
@@ -375,7 +445,7 @@ class BaseIntervalInfoModel(models.Model):
     info_date_next = models.DateTimeField(db_index = True, default = datetime.max)
     info_date_begin = models.DateTimeField(db_index = True, blank = True)
     info_date_end = models.DateTimeField(db_index = True, blank = True)
-    
+
     dimentions = [] # перечень реквизитов-измерений
     period = PERIOD_DAY # тип периодичности
 
@@ -391,6 +461,11 @@ class BaseIntervalInfoModel(models.Model):
     # мы сами берем на себя отвественность
     # за настройку полей(тип, индексы, default, allow_blank и т.д.)
 
+    def __init__(self, *args, **kwargs):
+        super(BaseIntervalInfoModel, self).__init__(*args, **kwargs)
+        # словарь параметров для передачи через сигналы
+        self._signal_params = {}
+
     def _get_safe_value(self, fld_name):
         u"""
             Возвращает значение для поля.
@@ -422,29 +497,43 @@ class BaseIntervalInfoModel(models.Model):
     # сформируем запрос по ключу
     @classmethod
     def query_dimentions(cls, data):
-        '''
-        запрос по ключевым полям
-        @data - объект, который содержит данные для сохранения в регистр.
-                Это может быть либо 1) объект модели хранения данных регистра,
-                либо 2) словарь, ключами которого являются имена полей из модели хранения,
-                либо 3) объект любого класса, у которого есть атрибуты с именами
-        '''
-        # добавим в запрос фильтр по ключевым полям, со значениями = текущему объекту, переданным данным
+        """ Запрос по ключевым полям.
+
+        :param data: объект, который содержит данные для сохранения в
+            регистр. Это может быть:
+            - Либо объект модели хранения данных регистра,
+            - Либо словарь, ключами которого являются имена полей из
+              модели хранения,
+            - Либо объект любого класса, у которого есть атрибуты с
+              именами
+        """
+        # Добавим в запрос фильтр по ключевым полям, со значениями =
+        # текущему объекту, переданным данным
         query = cls.objects
         for dim_field in cls.dimentions:
             dim_val = None
             dim_attr = dim_field
+
             if isinstance(dim_field, models.Field):
                 dim_attr = dim_field.name
+
             if isinstance(data, BaseInfoModel):
+                if not hasattr(data, dim_attr):
+                    continue
                 dim_val = getattr(data, dim_attr, None)
+
             elif isinstance(data, dict):
-                if data.has_key(dim_attr):
-                    dim_val = data[dim_attr]
+                if dim_attr not in data:
+                    continue
+                dim_val = data[dim_attr]
+
             elif hasattr(data, dim_attr):
+                if not hasattr(data, dim_attr):
+                    continue
                 dim_val = getattr(data, dim_attr, None)
-            if dim_val:
-                query = query.filter(**{dim_attr: dim_val})
+
+            query = query.filter(**{dim_attr: dim_val})
+
         return query
 
     @classmethod
@@ -467,18 +556,18 @@ class BaseIntervalInfoModel(models.Model):
         elif src == 'end':
             return Q(**{_tmpl % (cls.end_src, add): val})
         raise RuntimeError('invalid arg <<src>> for _gen_Q method!')
-    
+
     # сформируем запрос на интервал дат
-    @classmethod 
-    def query_interval(cls, data, date_begin = datetime.min, date_end = datetime.max, include_begin = True, include_end = True):
-        '''
+    @classmethod
+    def query_interval(cls, data, date_begin=datetime.min, date_end=datetime.max, include_begin=True, include_end=True):
+        u"""
         Выборка записей, попадающий в указанный интервал дат.
         @data - объект или словарь с ключевыми данными
         @date_begin - начало интервала
         @date_end - конец интервала
         @include_begin = True - запись учитывается если начало интервала попадает в интервал записи, иначе не учитывается
         @include_end = True - запись учитывается если конец интервала попадает в интервал записи, иначе не учитывается
-        '''
+        """
         query = cls.query_dimentions(data)
         if cls.period != PERIOD_INFTY:
             date_begin = normdate(cls.period, date_begin, True)
@@ -490,17 +579,17 @@ class BaseIntervalInfoModel(models.Model):
                     filter = filter | (cls._gen_Q('bgn', 'lt', date_end) & cls._gen_Q('end', 'gte', date_end)) # попадание конца интервала в границы записи
         query = query.filter(filter)
         return query
-    
+
     # сформируем запрос данных на дату
     @classmethod
     def query_on_date(cls, data, date=None, active=True, next=False):
-        '''
+        u"""
         запрос записей на дату
         @data - объект или словарь с ключевыми данными
         @date - дата актуальности, по-умолчанию текущая
         @active = True - значит будут найдены записи, действующие на указанную дату
         @next = False - значит будут найдены записи, действовавшие до даты, иначе ближайшиие записи которые будут действовать
-        '''
+        """
         if date is None:
             date = datetime.now()
 
@@ -516,7 +605,7 @@ class BaseIntervalInfoModel(models.Model):
                 filter = filter | (cls._gen_Q('bgn', 'lte', q_date_begin) & cls._gen_Q('end', 'gte', q_date_end)) # дата попадает в интервал записи
             query = query.filter(filter)
         return query
-    
+
     # прямая запись объекта, чтобы можно было записывать без доп. обработки
     def _save(self, *args, **kwargs):
         super(BaseIntervalInfoModel, self).save(*args, **kwargs)
@@ -526,13 +615,13 @@ class BaseIntervalInfoModel(models.Model):
         # если не указана дата записи, то
         if self.period != PERIOD_INFTY and (not self._begin or not self._end):
             raise Exception('Attribute info_date is not set!')
-        # нормализуем переданные дату и время 
+        # нормализуем переданные дату и время
         self._begin = normdate(self.period,self._begin, True)
         self._end = normdate(self.period,self._end, False)
 
         if self.period != PERIOD_INFTY and (self._begin > self._end):
             raise InvalidIntervalError("%s > %s"%(self.__class__.begin_src, self.__class__.end_src))
-        
+
         # проверим на уникальность записи
         if self.period != PERIOD_INFTY:
             q = self.__class__.query_interval(self, self._begin, self._end)
@@ -540,7 +629,7 @@ class BaseIntervalInfoModel(models.Model):
                 q = q.exclude(pk = self.pk)
             if q.count() > 0:
                 raise OverlapError( self.__class__ )
-        
+
         # если объект уже хранится, то найдем записи для изменения
         old_prev_rec = None
         old_next_rec = None
@@ -585,48 +674,29 @@ class BaseIntervalInfoModel(models.Model):
                 new_next_rec._save()
 
     # прямое удаление, без обработки
+    @InfoModelMeta.suppress_signals
     def _delete(self, *args, **kwargs):
         super(BaseIntervalInfoModel, self).delete(*args, **kwargs)
 
-    # переопределим удаление объекта, чтобы изменять соседние записи
-    def delete(self, *args, **kwargs):
-        # если объект уже хранится, то найдем записи для изменения
-        if self.pk:
-            q = self.__class__.query_dimentions(self).filter(self.__class__._gen_Q('end','',self.info_date_prev)).exclude(pk = self.pk)
-            old_prev_rec = q.get() if len(q) == 1 else None
-            q = self.__class__.query_dimentions(self).filter(self.__class__._gen_Q('bgn', '', self.info_date_next)).exclude(pk = self.pk)
-            old_next_rec = q.get() if len(q) == 1 else None
-        else:
-            old_prev_rec = None
-            old_next_rec = None
-        # удалим запись
-        self._delete(*args, **kwargs)
-        # изменим найденные ранее записи
-        #TODO: тут можно оптимизировать сохранение, чтобы небыло каскада сохранений
-        if old_prev_rec:
-            old_prev_rec.save()
-        if old_next_rec:
-            old_next_rec.save()
-    
     class Meta:
         abstract = True
 
 
-def RebuildIntervalInfoModel(cls, on_overlap_error = 0):
-    '''
+def RebuildIntervalInfoModel(cls, on_overlap_error=0):
+    u"""
     Перестраивает связи между записями в интервальном регистре
     on_overlap_error: действие при возникновении ошибки перекрытия диапазонов записей
     0 - вызвать exception
     1 - изменить интервал записи с ранней датой начала
     2 - выдать список ID глючных элементов, которые пропущены (после исправления придется еще раз запускать пересчет)
-    ''' 
-    
+    """
+
     def eq_keys(dims, key1, key2):
         for key_attr in dims:
             if key1[key_attr] != key2[key_attr]:
                 return False
         return True
-    
+
     def get_key(dims, data):
         key = {}
         for key_attr in dims:
@@ -634,14 +704,15 @@ def RebuildIntervalInfoModel(cls, on_overlap_error = 0):
             if isinstance(data, BaseIntervalInfoModel):
                 dim_val = getattr(data, key_attr, None)
             elif isinstance(data, dict):
-                if data.has_key(key_attr):
+                if key_attr in data:
                     dim_val = data[key_attr]
             elif hasattr(data, key_attr):
                 dim_val = getattr(data, key_attr, None)
             key[key_attr] = dim_val
         return key
-    
-    # вытащить записи регистра сгруппированные по ключевым полям и отсортированные по дате
+
+    # Вытащить записи регистра сгруппированные по ключевым полям и
+    # отсортированные по дате
     query = cls.objects
     order = []
     dims = []
@@ -686,7 +757,7 @@ def RebuildIntervalInfoModel(cls, on_overlap_error = 0):
                         continue
                     else:
                         raise OverlapError()
-            
+
             last_rec.info_date_next = rec._begin
             last_rec._save()
             rec.info_date_prev = last_rec._end
